#from __future__ import print_function, division

import torch
import numpy as np
from sklearn.preprocessing import StandardScaler
import random
from PIL import Image
import torch.utils.data as data
import os
import os.path
import accimage

def make_dataset(image_list, labels, root):
    if labels:  # labels=None for imagenet
      len_ = len(image_list)
      images = [(image_list[i].strip(), labels[i, :]) for i in range(len_)]
    else:      # split and get the labels
      if len(image_list[0].split()) > 2:
        images = [(os.path.join(root, val.split()[0]), np.array([int(la) for la in val.split()[1:]])) for val in image_list]
      else:
        images = [(val.split()[0], int(val.split()[1])) for val in image_list]
    return images


def pil_loader(path):
    # open path as file to avoid ResourceWarning (https://github.com/python-pillow/Pillow/issues/835)
    with open(path, 'rb') as f:
        with Image.open(f) as img:
            return img.convert('RGB')


def accimage_loader(path):
    import accimage
    try:
        return accimage.Image(path)
    except IOError:
        # Potentially a decoding problem, fall back to PIL.Image
        return pil_loader(path)


def default_loader(path):
    #from torchvision import get_image_backend
    #if get_image_backend() == 'accimage':
    #    return accimage_loader(path)
    #else:
        return pil_loader(path)


class ImageList(object):
    """A generic data loader where the images are arranged in this way: ::
        root/dog/xxx.png
        root/dog/xxy.png
        root/dog/xxz.png
        root/cat/123.png
        root/cat/nsdf3.png
        root/cat/asd932_.png
    Args:
        root (string): Root directory path.
        transform (callable, optional): A function/transform that  takes in an PIL image
            and returns a transformed version. E.g, ``transforms.RandomCrop``
        target_transform (callable, optional): A function/transform that takes in the
            target and transforms it.
        loader (callable, optional): A function to load an image given its path.
     Attributes:
        classes (list): List of the class names.
        class_to_idx (dict): Dict with items (class_name, class_index).
        imgs (list): List of (image path, class_index) tuples
    """
    train_list = 'train.txt'
    db_list = 'database.txt'
    query_list = 'test.txt'

    def __init__(self, root, split='train', labels=None, transform=None, target_transform=None,
                 loader=default_loader):  # ImageList(image_list = '../data/imagenet/train.txt')
        self.root = os.path.expanduser(root)
        if split == 'train':
            image_list = open(os.path.join(self.root, self.train_list)).readlines()
        elif split == 'gallery':
            image_list = open(os.path.join(self.root, self.db_list)).readlines()
        elif split == 'query':
            image_list = open(os.path.join(self.root, self.query_list)).readlines()
        else:
            assert False

        imgs = make_dataset(image_list, labels, self.root)

        if len(imgs) == 0:
            raise(RuntimeError("Found 0 images in subfolders of: " + root + "\n"
                               "Supported image extensions are: " + ",".join(IMG_EXTENSIONS)))

        self.imgs = imgs
        self.transform = transform
        self.target_transform = target_transform
        self.loader = loader
        self.nb_classes = 21

        if split == 'query':
            db_imgs_list = open(os.path.join(self.root, self.db_list)).readlines()
            db_imgs = make_dataset(db_imgs_list, labels, self.root)
            gallery_one_hot = np.array(list(zip(*db_imgs))[1])

            query_one_hot = np.array(list(zip(*imgs))[1])
            self.label_mat = (np.matmul(query_one_hot, np.transpose(gallery_one_hot)) > 0).astype(np.float32)
            #print(self.label_mat)

    def __getitem__(self, index):
        """
        Args:
            index (int): Index
        Returns:
            tuple: (image, target) where target is class_index of the target class.
        """
        path, target = self.imgs[index]
        img = self.loader(path)
        if self.transform is not None:
            img = self.transform(img)
        if self.target_transform is not None:
            target = self.target_transform(target)

        return img, target

    def __len__(self):
        return len(self.imgs)

if __name__ == '__main__':
    root = '../data/nus_wide'
    train_dataset = ImageList(root, split='train')

    for x in train_dataset:
        print(x)


